"""MySQL module (native)."""

from contextlib import contextmanager
from pathlib import Path
from typing import Any, Dict, Generator, Optional

from pymysql.connections import Connection

from spicerack.constants import PUPPET_CA_PATH
from spicerack.exceptions import SpicerackError


class MysqlError(SpicerackError):
    """Custom exception class for errors of this module."""


class Mysql:
    """Class to manage MySQL servers.

    Caution:
        This class only has DRY-RUN support for DML sql operations.

    """

    def __init__(self, *, dry_run: bool = True) -> None:
        """Initialize the instance.

        Arguments:
            dry_run (bool, optional): whether this is a DRY-RUN.

        """
        self._dry_run = dry_run

    @contextmanager
    def connect(
        self,
        *,
        read_only: bool = False,
        charset: str = "utf8mb4",
        read_default_file: Optional[str] = "",
        read_default_group: Optional[str] = None,
        ssl: Optional[Dict] = None,
        **kwargs: Any
    ) -> Generator:
        """Context-manager for a mysql connection to a remote host.

        Caution:
            Read-only support is limited to DML sql operations.

        Important:
            - This does not commit changes, that is the caller's responsibility.
            - The caller should also take care of rolling back transactions on error
              as appropriate.

        Arguments:
            read_only (bool, optional): True if this connection should use read-only
                transactions. **Note**: This parameter has no effect if DRY-RUN is set.
            charset (str, optional): Query charset to use.
            read_default_file (str, optional): ``my.cnf``-format file to read from. Defaults
                to ``~/.my.cnf``. Set to :py:data:`None` to disable.
            read_default_group: Section of read_default_file to use. If not specified, it
                will be set based on the target hostname.
            ssl (dict, optional): SSL configuration to use. Defaults to using the
                puppet CA. Set to ``{}`` to disable.
            **kwargs: Options passed directly to :py:class:`pymysql.connections.Connection`.

        Yields:
            :py:class:`pymysql.connections.Connection`: a context-managed mysql connection.

        """
        # FIXME(kormat): read-only support is limited to DML sql statements.
        # https://phabricator.wikimedia.org/T254756 is needed to do this better.
        read_only = read_only or self._dry_run

        if read_default_file == "":
            read_default_file = str(Path("~/.my.cnf").expanduser())
        if read_default_file and not read_default_group:
            read_default_group = "client"
            if kwargs.get("host", "").startswith("labsdb"):
                read_default_group += "labsdb"
        if ssl is None:
            ssl = {"ca": PUPPET_CA_PATH}

        conn = Connection(
            charset=charset,
            read_default_file=read_default_file,
            read_default_group=read_default_group,
            ssl=ssl,
            **kwargs,
        )
        if read_only:
            conn.query("SET SESSION TRANSACTION READ ONLY")
        try:
            yield conn
        # Not catching exceptions and rolling back, as that restricts the client code
        # in how it does error handling.
        finally:
            conn.close()



"""Alertmanager module tests."""

from unittest import mock

import pytest

from spicerack import alerting, alertmanager, icinga
from spicerack.administrative import Reason
from spicerack.remote import RemoteHosts
from spicerack.tests.unit.test_icinga import set_mocked_icinga_host_output


class TestAlerting:
    """Tests for the Alerting class."""

    @pytest.fixture(autouse=True)
    def setup_method(self, requests_mock):
        """Initialize the test instance."""
        # pylint: disable=attribute-defined-outside-init
        self.hosts = ["host1.example.com"]
        self.reason = Reason("test", "user", "host")

        self.mocked_icinga_host = mock.MagicMock(spec_set=RemoteHosts)
        self.mocked_icinga_host.__len__.return_value = 1
        set_mocked_icinga_host_output(self.mocked_icinga_host, "/var/lib/icinga/rw/icinga.cmd")
        self.icinga_hosts = icinga.IcingaHosts(self.mocked_icinga_host, self.hosts)

        self.am_hosts = alertmanager.AlertmanagerHosts(self.hosts)
        self.requests_mock = requests_mock

        self.alerting_hosts = alerting.AlertingHosts(self.am_hosts, self.icinga_hosts)

    @mock.patch("spicerack.icinga.IcingaHosts.downtime")
    @mock.patch("spicerack.alertmanager.AlertmanagerHosts.downtime", return_value="foo")
    def test_downtime(self, am_dt, icinga_dt):
        """It should call both Alertmanager and Icinga."""
        assert "foo" == self.alerting_hosts.downtime(self.reason)
        assert am_dt.called
        assert icinga_dt.called

    @mock.patch("spicerack.icinga.IcingaHosts.remove_downtime")
    @mock.patch("spicerack.alertmanager.AlertmanagerHosts.remove_downtime")
    def test_remove_downtime(self, am_rmdt, icinga_rmdt):
        """It should call both Alertmanager and Icinga."""
        self.alerting_hosts.remove_downtime("foo")
        assert am_rmdt.called
        assert icinga_rmdt.called

    @mock.patch("spicerack.icinga.IcingaHosts.downtime")
    @mock.patch("spicerack.alertmanager.AlertmanagerHosts.downtime")
    @mock.patch("spicerack.icinga.IcingaHosts.remove_downtime")
    @mock.patch("spicerack.alertmanager.AlertmanagerHosts.remove_downtime")
    def test_downtimed(self, am_rmdt, icinga_rmdt, am_dt, icinga_dt):
        """It should issue a silence and then delete it."""
        with self.alerting_hosts.downtimed(self.reason):
            assert am_dt.called
            assert icinga_dt.called
        assert am_rmdt.called
        assert icinga_rmdt.called

    @pytest.mark.parametrize(
        "remove_on_error, remove_downtime_calls",
        (
            (True, 1),
            (False, 0),
        ),
    )
    @mock.patch("spicerack.icinga.IcingaHosts.downtime")
    @mock.patch("spicerack.alertmanager.AlertmanagerHosts.downtime")
    @mock.patch("spicerack.icinga.IcingaHosts.remove_downtime")
    @mock.patch("spicerack.alertmanager.AlertmanagerHosts.remove_downtime")
    def test_downtimed_on_error(
        self, am_rmdt, icinga_rmdt, am_dt, icinga_dt, remove_on_error, remove_downtime_calls
    ):  # pylint: disable=too-many-arguments
        """It should issue a silence and then delete it even with errors."""
        with pytest.raises(ValueError):
            with self.alerting_hosts.downtimed(self.reason, remove_on_error=remove_on_error):
                assert am_dt.called
                assert icinga_dt.called
                raise ValueError()
        assert am_rmdt.call_count == remove_downtime_calls
        assert icinga_rmdt.call_count == remove_downtime_calls



"""Redis cluster module."""
import logging
from collections import defaultdict
from pathlib import Path
from typing import Any, DefaultDict, Dict, Tuple, Union

from redis import StrictRedis
from wmflib.config import load_yaml_config

from spicerack.exceptions import SpicerackError

logger = logging.getLogger(__name__)


class RedisClusterError(SpicerackError):
    """Custom exception class for errors in the RedisCluster class."""


class RedisCluster:
    """Class to manage a Redis Cluster."""

    def __init__(self, cluster: str, config_dir: Path, *, dry_run: bool = True) -> None:
        """Initialize the instance.

        Arguments:
            cluster (str): the name of the cluster to connect to.
            config_dir (str): path to the directory containing the configuration files for the Redis clusters.
            dry_run (bool, optional): whether this is a DRY-RUN.

        """
        self._dry_run = dry_run
        self._shards: DefaultDict[str, Dict] = defaultdict(dict)
        config = load_yaml_config(config_dir / f"{cluster}.yaml")

        for datacenter, shards in config["shards"].items():
            for shard, data in shards.items():
                self._shards[datacenter][shard] = RedisInstance(
                    host=data["host"],
                    port=data["port"],
                    password=config["password"],
                    decode_responses=True,
                )

    def start_replica(self, datacenter: str, master_datacenter: str) -> None:
        """Start the cluster replica in a datacenter from a master datacenter.

        Arguments:
            datacenter (str): the datacenter on which to start the replica.
            master_datacenter (str): the datacenter from which to replicate.

        Raises:
            RedisClusterError: on error and invalid parameters.

        """
        if master_datacenter == datacenter:
            raise RedisClusterError(
                f"Master datacenter must be different from the current datacenter, got {datacenter}"
            )

        for shard, instance in sorted(self._shards[datacenter].items()):
            self._start_instance_replica(instance, self._shards[master_datacenter][shard])

    def stop_replica(self, datacenter: str) -> None:
        """Stop the cluster replica in a datacenter.

        Arguments:
            datacenter (str): the datacenter on which to stop the replica.

        Raises:
            RedisClusterError: on error.

        """
        for instance in self._shards[datacenter].values():
            self._stop_instance_replica(instance)

    def _start_instance_replica(self, instance: "RedisInstance", master: "RedisInstance") -> None:
        """Start the replica in a specific instance from a master instance.

        Arguments:
            instance (spicerack.redis_cluster.RedisInstance): the instance where to start the replica.
            master (spicerack.redis_cluster.RedisInstance): the master instance to replicate from.

        Raises:
            RedisClusterError: if unable to verify the replica has started.

        """
        if instance.master_info == master.info:
            logger.debug("Replica already configured on %s", instance)
            return

        if self._dry_run:
            logger.debug("Skip starting replica on %s in dry-run mode", instance)
        else:
            logger.debug("Starting replica %s => %s", master, instance)
            instance.start_replica(master)

        if not self._dry_run and instance.master_info != master.info:
            raise RedisClusterError(f"Replica on {instance} is not correctly configured: {instance.master_info}")

    def _stop_instance_replica(self, instance: "RedisInstance") -> None:
        """Stop the replica in a specific instance.

        Arguments:
            instance (spicerack.redis_cluster.RedisInstance): the instance where to stop the replica.

        Raises:
            RedisClusterError: on error.

        """
        if instance.is_master:
            logger.debug("Instance %s is already master, doing nothing", instance)
            return

        if self._dry_run:
            logger.debug("Skip stopping replica on %s in dry-run mode", instance)
        else:
            logger.debug("Stopping replica on %s", instance)
            instance.stop_replica()

        if not self._dry_run and not instance.is_master:
            raise RedisClusterError(f"Instance {instance} is still a slave of {instance.master_info}, aborting")


class RedisInstance:
    """Class to manage a Redis instance, a simple wrapper around `redis.StrictRedis`."""

    def __init__(self, **kwargs: Any) -> None:
        """Initialize the instance.

        Arguments:
            **kwargs (mixed): arbitrary keyword arguments, to be passed to the `redis.StrictRedis` constructor.

        """
        self.host: str = kwargs.get("host", "")
        self.port: int = kwargs.get("port", 0)
        self._client = StrictRedis(**kwargs)

    @property
    def is_master(self) -> bool:
        """Getter to check if the current instance is a master.

        Returns:
            bool: :py:data:`True` if the instance is a master, :py:data:`False` otherwise.

        """
        return self._client.info("replication")["role"] == "master"

    @property
    def master_info(self) -> Union[Tuple[None, None], Tuple[str, int]]:
        """Getter to know the master of this instance.

        Returns:
            tuple: a 2-element tuple with (host/IP, port) of the master instance. If there is no master configured
            (None, None) is returned.

        """
        data = self._client.info("replication")
        try:
            return data["master_host"], data["master_port"]
        except (KeyError, TypeError):
            return (None, None)

    @property
    def info(self) -> Tuple[str, int]:
        """Getter to know the detail of this instance.

        Returns:
            tuple: a 2-element tuple with (host/IP, port) of the instance.

        """
        return self.host, self.port

    def stop_replica(self) -> None:
        """Stop the replica on the instance."""
        self._client.slaveof()

    def start_replica(self, master: "RedisInstance") -> None:
        """Start the replica from the given master instance.

        Arguments:
            master (spicerack.redis_cluster.RedisInstance): the master instance.

        """
        self._client.slaveof(master.host, master.port)

    def __str__(self) -> str:
        """String representation of the instance.

        Returns:
            str: the host or IP and port of the instance.

        """
        return f"{self.host}:{self.port}"



"""Debmonitor module."""
import logging

import requests
from wmflib.requests import http_session

from spicerack.exceptions import SpicerackError

logger = logging.getLogger(__name__)


class DebmonitorError(SpicerackError):
    """Custom exception class for errors of the Debmonitor class."""


class Debmonitor:
    """Class to interact with a Debmonitor website."""

    def __init__(self, host: str, cert: str, key: str, dry_run: bool = True) -> None:
        """Initialize the instance.

        Arguments:
            host (str): the hostname of the Debmonitor server (without protocol).
            cert (str): the path to the TLS certificate to use to authenticate on Debmonitor.
            key (str): the path to the TLS key to use to authenticate on Debmonitor.
            dry_run (bool, optional): whether this is a DRY-RUN.

        """
        self._base_url: str = f"https://{host}"
        self._cert = cert
        self._key = key
        self._dry_run = dry_run
        self._http_session = http_session(".".join((self.__module__, self.__class__.__name__)))

    def host_delete(self, hostname: str) -> None:
        """Remove a host and all its packages from Debmonitor.

        Arguments:
            host (str): the FQDN of the host to remove from Debmonitor.

        Raises:
            spicerack.debmonitor.DebmonitorError: on failure to delete. It doesn't raise if the host is already absent
                in Debmonitor.

        """
        if self._dry_run:
            logger.debug("Skip removing host %s from Debmonitor in DRY-RUN", hostname)
            return

        url = f"{self._base_url}/hosts/{hostname}"
        response = self._http_session.delete(url, cert=(self._cert, self._key))

        if response.status_code == requests.codes["no_content"]:
            logger.info("Removed host %s from Debmonitor", hostname)
        elif response.status_code == requests.codes["not_found"]:
            logger.info("Host %s already missing on Debmonitor", hostname)
        else:
            raise DebmonitorError(
                f"Unable to remove host {hostname} from Debmonitor, got: {response.status_code} {response.reason}"
            )



"""RedisCluster module tests."""
from unittest import mock

import pytest

from spicerack.redis_cluster import RedisCluster, RedisClusterError
from spicerack.tests import get_fixture_path


class TestRedisCluster:
    """RedisCluster class tests."""

    @mock.patch("spicerack.redis_cluster.StrictRedis")
    def setup_method(self, _, mocked_redis):
        """Initialize the test environment for RedisCluster."""
        config_dir = get_fixture_path("redis_cluster")
        # pylint: disable=attribute-defined-outside-init
        self.mocked_redis = mocked_redis
        self.redis_cluster = RedisCluster("cluster", config_dir, dry_run=False)
        self.redis_cluster_dry_run = RedisCluster("cluster", config_dir)

    def test_start_replica(self):
        """Should start the replica and verify it."""
        self.mocked_redis.return_value.info.side_effect = [
            {"role": "master"},
            {"role": "slave", "master_host": "host1", "master_port": 123},
            {"role": "master"},
            {"role": "slave", "master_host": "host2", "master_port": 123},
        ]
        self.redis_cluster.start_replica("dc2", "dc1")
        assert self.mocked_redis.return_value.slaveof.called

    def test_start_replica_dry_run(self):
        """Should skip starting the replica in dry-run mode and not raise exception."""
        self.mocked_redis.return_value.info.side_effect = [{"role": "master"}] * 2
        self.redis_cluster_dry_run.start_replica("dc2", "dc1")
        assert not self.mocked_redis.return_value.slaveof.called

    def test_start_replica_noop(self):
        """Should be a noop if the replica is already correctly configured."""
        self.mocked_redis.return_value.info.side_effect = [
            {"role": "slave", "master_host": "host1", "master_port": 123},
            {"role": "slave", "master_host": "host2", "master_port": 123},
        ]
        self.redis_cluster.start_replica("dc2", "dc1")
        assert not self.mocked_redis.return_value.slaveof.called

    def test_start_replica_fail(self):
        """Should raise RedisClusterError if not able to verify that is started."""
        self.mocked_redis.return_value.info.side_effect = [{"role": "master"}] * 3
        with pytest.raises(RedisClusterError, match="is not correctly configured"):
            self.redis_cluster.start_replica("dc2", "dc1")

    def test_start_replica_same_dc(self):
        """Should raise RedisClusterError when trying to set the replica with it's own datacenter."""
        with pytest.raises(
            RedisClusterError,
            match="Master datacenter must be different from the current datacenter",
        ):
            self.redis_cluster.start_replica("dc1", "dc1")

        assert not self.mocked_redis.return_value.info.called

    def test_stop_replica(self):
        """Should stop the replica and verify it."""
        self.mocked_redis.return_value.info.side_effect = [
            {"role": "slave", "master_host": "host1", "master_port": 123},
            {"role": "master"},
        ] * 2
        self.redis_cluster.stop_replica("dc2")
        assert self.mocked_redis.return_value.slaveof.called

    def test_stop_replica_dry_run(self):
        """Should stop the replica and verify it."""
        self.mocked_redis.return_value.info.side_effect = [
            {"role": "slave", "master_host": "host1", "master_port": 123},
            {"role": "master"},
        ] * 2
        self.redis_cluster_dry_run.stop_replica("dc2")
        assert not self.mocked_redis.return_value.slaveof.called

    def test_stop_replica_noop(self):
        """Should skip stopping the replica in dry-run mode and not raise exception."""
        self.mocked_redis.return_value.info.side_effect = [{"role": "master"}] * 2
        self.redis_cluster.stop_replica("dc2")
        assert not self.mocked_redis.return_value.slaveof.called

    def test_stop_replica_fail(self):
        """Should raise RedisClusterError if not able to verify that is stopped."""
        self.mocked_redis.return_value.info.side_effect = [
            {"role": "slave", "master_host": "host1", "master_port": 123}
        ] * 3
        with pytest.raises(RedisClusterError, match="is still a slave of"):
            self.redis_cluster.stop_replica("dc2")



"""Spicerack package."""
from logging import Logger, getLogger
from pathlib import Path
from socket import gethostname
from typing import TYPE_CHECKING, Callable, Dict, Optional, Sequence

from git import Repo
from pkg_resources import DistributionNotFound, get_distribution
from wmflib import requests
from wmflib.actions import ActionsDict
from wmflib.config import load_ini_config, load_yaml_config
from wmflib.dns import Dns
from wmflib.interactive import get_username
from wmflib.phabricator import Phabricator, create_phabricator
from wmflib.prometheus import Prometheus

from spicerack._log import irc_logger
from spicerack.administrative import Reason
from spicerack.alerting import AlertingHosts
from spicerack.alertmanager import AlertmanagerHosts
from spicerack.confctl import Confctl, ConftoolEntity
from spicerack.debmonitor import Debmonitor
from spicerack.dhcp import DHCP
from spicerack.dnsdisc import Discovery
from spicerack.elasticsearch_cluster import ElasticsearchClusters, create_elasticsearch_clusters
from spicerack.exceptions import SpicerackError
from spicerack.ganeti import Ganeti
from spicerack.icinga import ICINGA_DOMAIN, IcingaHosts
from spicerack.interactive import get_management_password
from spicerack.ipmi import Ipmi
from spicerack.kafka import Kafka
from spicerack.mediawiki import MediaWiki
from spicerack.mysql import Mysql
from spicerack.mysql_legacy import MysqlLegacy
from spicerack.netbox import NETBOX_DOMAIN, Netbox, NetboxServer
from spicerack.puppet import PuppetHosts, PuppetMaster, get_puppet_ca_hostname
from spicerack.redfish import Redfish, RedfishDell
from spicerack.redis_cluster import RedisCluster
from spicerack.remote import Remote, RemoteHosts
from spicerack.reposync import RepoSync
from spicerack.toolforge.etcdctl import EtcdctlController
from spicerack.typing import TypeHosts

if TYPE_CHECKING:  # Imported only during type checking, prevents cyclic imports at runtime
    from spicerack._menu import BaseItem  # pragma: no cover


logger = getLogger(__name__)

try:
    __version__: str = get_distribution("wikimedia-spicerack").version  # Must be the same used as 'name' in setup.py
    """:py:class:`str`: the version of the current Spicerack module."""
except DistributionNotFound:  # pragma: no cover - this should never happen during tests
    pass  # package is not installed


class Spicerack:  # pylint: disable=too-many-instance-attributes
    """Spicerack service locator."""

    def __init__(
        self,
        *,
        verbose: bool = False,
        dry_run: bool = True,
        cumin_config: str = "/etc/cumin/config.yaml",
        cumin_installer_config: str = "/etc/cumin/config-installer.yaml",
        conftool_config: str = "/etc/conftool/config.yaml",
        conftool_schema: str = "/etc/conftool/schema.yaml",
        debmonitor_config: str = "/etc/debmonitor.conf",
        spicerack_config_dir: str = "/etc/spicerack",
        http_proxy: str = "",
        get_cookbook_callback: Optional[Callable[["Spicerack", str, Sequence[str]], Optional["BaseItem"]]] = None,
    ) -> None:
        """Initialize the service locator for the Spicerack library.

        Arguments:
            verbose (bool, optional): whether to set the verbose mode.
            dry_run (bool, optional): whether this is a DRY-RUN.
            cumin_config (str, optional): the path to Cumin's configuration file.
            cumin_installer_config (str, optional): the path to Cumin's configuration file for the Debian installer.
            conftool_config (str, optional): the path to Conftool's configuration file.
            conftool_schema (str, optional): the path to Conftool's schema file.
            debmonitor_config (str, optional): the path to Debmonitor's INI configuration file. It must have at least
                the following schema::

                    [DEFAULT]
                    server=debmonitor.example.com
                    cert=/etc/debmonitor/ssl/cert.pem
                    key=/etc/debmonitor/ssl/server.key

            spicerack_config_dir (str, optional): the path for the root configuration directory for Spicerack.
                Module-specific configuration will be loaded from `config_dir/module_name/`.
            http_proxy (str, optional): the scheme://url:port of the HTTP proxy to use for external calls.
            get_cookbook_callback (callable, optional): a callable to retrieve a CookbookItem to execute a cookbook
                from inside another cookbook.

        """
        # Attributes
        self._verbose = verbose
        self._dry_run = dry_run
        self._http_proxy = http_proxy
        self._cumin_config = cumin_config
        self._cumin_installer_config = cumin_installer_config
        self._conftool_config = conftool_config
        self._conftool_schema = conftool_schema
        self._debmonitor_config = debmonitor_config
        self._spicerack_config_dir = Path(spicerack_config_dir)
        self._get_cookbook_callback = get_cookbook_callback

        self._username = get_username()
        self._current_hostname = gethostname()
        self._irc_logger = irc_logger
        self._confctl: Optional[Confctl] = None
        self._management_password: str = ""
        self._actions = ActionsDict()

    @property
    def dry_run(self) -> bool:
        """Getter for the ``dry_run`` property.

        Returns:
            bool: True if the DRY-RUN mode is set, False otherwise.

        """
        return self._dry_run

    @property
    def verbose(self) -> bool:
        """Getter for the ``verbose`` property.

        Returns:
            bool: True if the verbose mode is set, False otherwise.

        """
        return self._verbose

    @property
    def username(self) -> str:
        """Getter for the current username.

        Returns:
            str: the name of the effective running user.

        """
        return self._username

    @property
    def config_dir(self) -> Path:
        """Getter for Spicerack's configuration file directory.

        Returns:
            str: a filesystem location of configuration files.

        """
        return self._spicerack_config_dir

    @property
    def http_proxy(self) -> str:
        """Getter for the HTTP PROXY to use for external calls.

        Returns:
            str: the scheme://url:port of the proxy.

        """
        return self._http_proxy

    @property
    def requests_proxies(self) -> Optional[Dict[str, str]]:
        """Getter to return the HTTP proxy configuration for the Requests module.

        Returns:
            dict: with the proxies as required by Requests documentation.
            :py:data:`None`: if no HTTP proxy is set.

        See Also:
            https://requests.readthedocs.io/en/master/user/advanced/#proxies

        """
        if not self._http_proxy:
            return None

        return {"http": self._http_proxy, "https": self._http_proxy}

    @property
    def irc_logger(self) -> Logger:
        """Getter for the ``irc_logger`` property.

        Returns:
            logging.Logger: the logger instance to write to IRC.

        """
        return self._irc_logger

    @property
    def actions(self) -> ActionsDict:
        """Getter for the ``actions`` property.

        Returns:
            wmflib.actions.ActionsDict: a dictionary to log and record cookbook actions.

        """
        return self._actions

    @property
    def icinga_master_host(self) -> RemoteHosts:
        """Getter for the ``icinga_master_host`` property.

        Returns:
            spicerack.remote.RemoteHosts: the instance to execute commands on the Icinga master host.

        """
        return self.remote().query(self.dns().resolve_cname(ICINGA_DOMAIN))

    @property
    def netbox_master_host(self) -> RemoteHosts:
        """Getter for the ``netbox_master_host`` property.

        Returns:
            spicerack.remote.RemoteHosts: the instance to execute commands on the Netbox master host.

        """
        return self.remote().query(self.dns().resolve_cname(NETBOX_DOMAIN))

    @property
    def management_password(self) -> str:
        """Getter for the ``management_password`` property.

        Returns:
            str: the management password. It will be asked to the user if not already cached by the current instance.

        """
        if not self._management_password:
            self._management_password = get_management_password()

        return self._management_password

    def run_cookbook(self, cookbook: str, args: Sequence[str] = ()) -> int:
        """Run another Cookbook within the current run.

        The other Cookbook will be executed with the current setup and will log in the same file of the current
        Cookbook that is running.

        Arguments:
            cookbook (str): the path to the cookbook to execute, either in Spicerack's dot notation or the relative
                path to the Python file to execute.
            args (sequence, optional): an iterable sequence of strings with the Cookbook's argument. The Cookbook will
                be executed with the same global arguments used for the current run.

        Returns:
            int: the exit code of the Cookbook, 0 if successful, non-zero if not.

        Raises:
            spicerack.exceptions.SpicerackError: if the get_cookbook_callback callback is not set or unable to find the
                cookbook with the given name.

        """
        if self._get_cookbook_callback is None:
            raise SpicerackError("Unable to run other cookbooks, get_cookbook_callback is not set.")

        cookbook_item = self._get_cookbook_callback(self, cookbook, args)
        if cookbook_item is None:
            raise SpicerackError(f"Unable to find cookbook {cookbook}")

        logger.debug("Executing cookbook %s with args: %s", cookbook, args)
        return cookbook_item.run() or 0  # Force the return code to be 0 if the cookbook returns None

    def remote(self, installer: bool = False) -> Remote:
        """Get a Remote instance.

        Arguments:
            installer (bool, optional): whether to use the special configuration to connect to a Debian installer
                or freshly re-imaged host prior to its first Puppet run.

        Returns:
            spicerack.remote.Remote: the Remote instance.

        """
        return Remote(self._cumin_installer_config if installer else self._cumin_config, dry_run=self._dry_run)

    def confctl(self, entity_name: str) -> ConftoolEntity:
        """Access a Conftool specific entity instance.

        Arguments:
            entity_name (str): the name of a Conftool entity. Possible values: ``node``, ``service``, ``discovery``,
                ``mwconfig``.

        Returns:
            spicerack.confctl.ConftoolEntity: the confctl entity instance.

        """
        if self._confctl is None:
            self._confctl = Confctl(
                config=self._conftool_config,
                schema=self._conftool_schema,
                dry_run=self._dry_run,
            )

        return self._confctl.entity(entity_name)

    def dhcp(self, remote_hosts: RemoteHosts) -> DHCP:
        """Return a DHCP configuration manager for specified site.

        Arguments:
            remote_hosts (RemoteHosts instance): Hosts to operate on, which are normally install servers with dhcp.

        Returns:
            spicerack.dhcp.DHCP : A DHCP configuration instance

        """
        return DHCP(remote_hosts, dry_run=self._dry_run)

    def dns(self) -> Dns:  # pylint: disable=no-self-use
        """Get a Dns instance.

        Returns:
            wmflib.dns.Dns: a Dns instance that will use the operating system default namserver(s).

        """
        return Dns()

    def discovery(self, *records: str) -> Discovery:
        """Get a Discovery instance.

        Arguments:
            *records (str): arbitrary positional arguments, each one must be a Discovery DNS record name.

        Returns:
            spicerack.dnsdisc.Discovery: the pre-configured Discovery instance for the given records.

        """
        return Discovery(
            self.confctl("discovery"),
            self.remote(),
            list(records),
            dry_run=self._dry_run,
        )

    def mediawiki(self) -> MediaWiki:
        """Get a MediaWiki instance.

        Returns:
            spicerack.mediawiki.MediaWiki: the MediaWiki instance.

        """
        return MediaWiki(
            self.confctl("mwconfig"),
            self.remote(),
            self._username,
            dry_run=self._dry_run,
        )

    def mysql(self) -> Mysql:
        """Get a Mysql instance.

        Returns:
            spicerack.mysql.Mysql: the Mysql instance.

        """
        return Mysql(dry_run=self._dry_run)

    def mysql_legacy(self) -> MysqlLegacy:
        """Get a MysqlLegacy instance.

        Returns:
            spicerack.mysql_legacy.MysqlLegacy: the MysqlLegacy instance.

        """
        return MysqlLegacy(self.remote(), dry_run=self._dry_run)

    def redis_cluster(self, cluster: str) -> RedisCluster:
        """Get a RedisCluster instance.

        Arguments:
            cluster (str): the name of the cluster.

        Returns:
            spicerack.redis_cluster.RedisCluster: the cluster instance.

        """
        return RedisCluster(
            cluster,
            self._spicerack_config_dir / "redis_cluster",
            dry_run=self._dry_run,
        )

    def reposync(self, name: str) -> RepoSync:
        """Get a Reposync instance.

        Arguments:
            name (str): the name of the repo to sync.

        Returns:
            spicerack.reposync.RepoSync: the reposync instance.

        """
        config = load_yaml_config(self._spicerack_config_dir / "reposync" / "config.yaml")
        if name not in config["repos"]:
            raise SpicerackError(f"Unknown repo {name}")

        repo_dir = Path(config["base_dir"], name)
        query = ",".join(config["remotes"])
        remote_hosts = self.remote().query(query)

        if not repo_dir.is_dir():
            raise SpicerackError(f"The repo directory ({repo_dir}) does not exist")
        repo = Repo(repo_dir)
        if not repo.bare:
            raise SpicerackError(f"The repo directory ({repo_dir}) is not a bare git repository")

        return RepoSync(repo, self.username, remote_hosts, dry_run=self._dry_run)

    def elasticsearch_clusters(
        self, clustergroup: str, write_queue_datacenters: Sequence[str]
    ) -> ElasticsearchClusters:
        """Get an ElasticsearchClusters instance.

        Arguments:
            clustergroup (str): name of cluster group e.g ``search_eqiad``.
            write_queue_datacenters (Sequence[str]): Sequence of which core DCs to query write queues for.

        Returns:
            spicerack.elasticsearch_cluster.ElasticsearchClusters: ElasticsearchClusters instance.

        """
        configuration = load_yaml_config(self._spicerack_config_dir / "elasticsearch" / "config.yaml")

        return create_elasticsearch_clusters(
            configuration,
            clustergroup,
            write_queue_datacenters,
            self.remote(),
            self.prometheus(),
            dry_run=self._dry_run,
        )

    def admin_reason(self, reason: str, task_id: Optional[str] = None) -> Reason:
        """Get an administrative Reason instance.

        Arguments:
            reason (str): the reason to use to justify an administrative action. See `spicerack.administrative.Reason`
                for all the details.
            task_id (str, optional): the task ID to mention in the reason.

        Returns:
            spicerack.administrative.Reason: the administrative Reason instance.

        """
        return Reason(reason, self._username, self._current_hostname, task_id=task_id)

    def icinga_hosts(self, target_hosts: TypeHosts, *, verbatim_hosts: bool = False) -> IcingaHosts:
        """Get an IcingaHosts instance.

        Arguments:
            target_hosts (spicerack.typing.TypeHosts): the target hosts either as a NodeSet instance or a sequence
                of strings.
            verbatim_hosts (bool, optional): if :py:data:`True` use the hosts passed verbatim as is, if instead
                :py:data:`False`, the default, consider the given target hosts as FQDNs and extract their hostnames to
                be used in Icinga.

        Returns:
            spicerack.icinga.IcingaHosts: IcingaHosts instance.

        """
        return IcingaHosts(self.icinga_master_host, target_hosts, verbatim_hosts=verbatim_hosts)

    def puppet(self, remote_hosts: RemoteHosts) -> PuppetHosts:  # pylint: disable=no-self-use
        """Get a PuppetHosts instance for the given remote hosts.

        Arguments:
            remote_hosts (spicerack.remote.RemoteHosts): the instance with the target hosts.

        Returns:
            spicerack.puppet.PuppetHosts: the instance to manage Puppet on the target hosts.

        """
        return PuppetHosts(remote_hosts)

    def puppet_master(self) -> PuppetMaster:
        """Get a PuppetMaster instance to manage hosts and certificates from a Puppet master.

        Returns:
            spicerack.puppet.PuppetMaster: the instance to manage Puppet hosts and certificates.

        """
        return PuppetMaster(self.remote().query(get_puppet_ca_hostname()))

    def ipmi(self, mgmt_fqdn: str) -> Ipmi:
        """Get an Ipmi instance to send remote IPMI commands to management consoles.

        Arguments:
            mgmt_fqdn (str): the management console FQDN to target.

        Returns:
            spicerack.ipmi.Ipmi: the instance to run ipmitool commands.

        """
        return Ipmi(mgmt_fqdn, self.management_password, dry_run=self._dry_run)

    def phabricator(self, bot_config_file: str, section: str = "phabricator_bot") -> Phabricator:
        """Get a Phabricator instance to interact with a Phabricator website.

        Arguments:
            bot_config_file (str): the path to the configuration file for the Phabricator bot, with the following
                structure::

                    [section_name]
                    host = https://phabricator.example.com/api/
                    username = phab-bot
                    token = api-12345

            section (str, optional): the name of the section of the configuration file where to find the required
                parameters.

        Returns:
            wmflib.phabricator.Phabricator: the instance.

        """
        # Allow to specify the configuration file as opposed to other methods so that different clients can use
        # different Phabricator BOT accounts, potentially with different permissions.
        return create_phabricator(bot_config_file, section=section, dry_run=self._dry_run)

    def prometheus(self) -> Prometheus:  # pylint: disable=no-self-use
        """Get an Prometheus instance.

        Returns:
            wmflib.prometheus.Prometheus: Prometheus instance.

        """
        return Prometheus()

    def debmonitor(self) -> Debmonitor:
        """Get a Debmonitor instance to interact with a Debmonitor website.

        Returns:
            spicerack.debmonitor.Debmonitor: the instance.

        Raises:
            KeyError: if any configuration option is missing.

        """
        options = load_ini_config(self._debmonitor_config).defaults()
        return Debmonitor(options["server"], options["cert"], options["key"], dry_run=self._dry_run)

    def ganeti(self) -> Ganeti:
        """Get an instance to interact with Ganeti.

        Returns:
            spicerack.ganeti.Ganeti: the instance

        Raises:
            KeyError: If the configuration file does not contain the correct keys.

        """
        configuration = load_yaml_config(self._spicerack_config_dir / "ganeti" / "config.yaml")

        return Ganeti(
            configuration["username"],
            configuration["password"],
            configuration["timeout"],
            self.remote(),
        )

    def netbox(self, *, read_write: bool = False) -> Netbox:
        """Get a Netbox instance to interact with Netbox's API.

        Arguments:
            read_write (bool, optional): whether to use a read-write token.

        Returns:
            spicerack.netbox.Netbox: the instance

        """
        config = load_yaml_config(self._spicerack_config_dir / "netbox" / "config.yaml")
        if read_write and not self._dry_run:
            token = config["api_token_rw"]
        else:
            token = config["api_token_ro"]

        return Netbox(config["api_url"], token, dry_run=self._dry_run)

    def netbox_server(self, hostname: str, *, read_write: bool = False) -> NetboxServer:
        """Get a NetboxServer instance to interact with a server in Netbox, both physical and virtual.

        Arguments:
            hostname (str): the hostname (not FQDN) of the server to manage.
            read_write (bool, optional): whether to use a read-write token.

        Raises:
            spicerack.netbox.NetboxError: if unable to find or load the server.

        Returns:
            spicerack.netbox.NetboxServer: the NetboxServer instance.

        """
        return self.netbox(read_write=read_write).get_server(hostname)

    def requests_session(  # pylint: disable=no-self-use
        self,
        name: str,
        *,
        timeout: requests.TypeTimeout = requests.DEFAULT_TIMEOUT,
        tries: int = 3,
        backoff: float = 1.0,
    ) -> requests.Session:
        """Return a new requests Session with timeout and retry logic.

        Params:
            according to :py:func:`wmflib.requests.http_session`.

        Returns:
            requests.Session: the pre-configured session.

        """
        name = f"Spicerack/{__version__} {name}"
        return requests.http_session(name, timeout=timeout, tries=tries, backoff=backoff)

    def etcdctl(self, *, remote_host: RemoteHosts) -> EtcdctlController:  # pylint: disable=no-self-use
        """Add etcdctl control capabilities to the given RemoteHost.

        Params:
            remote_host: Single remote host (that should map one single host)
                to add capabilities to.

        Returns:
            A wrapped RemoteHost with the etcdctl control related methods.

        """
        return EtcdctlController(remote_host=remote_host)

    def kafka(self) -> Kafka:
        """Get an instance to interact with Kafka.

        Returns:
            spicerack.kafka.Kafka: the instance

        Raises:
            KeyError: If the configuration file does not contain the correct keys.

        """
        configuration = load_yaml_config(self._spicerack_config_dir / "kafka" / "config.yaml")

        return Kafka(kafka_config=configuration, dry_run=self._dry_run)

    def redfish(self, mgmt_fqdn: str, username: str, password: str = "") -> Redfish:  # nosec
        """Get an instance to talk to the Redfish API of a physical server.

        Notes:
            At the moment only Dell hardware is supported.

        Arguments:
            mgmt_fqdn (str): the management console FQDN to target.
            username (str): the username for the management console (usually 'root').
            password (str, optional): the password for the management console for the given user. If empty or not
                provided would use the production management password and ask the user for it if not already in memory.

        Returns:
            spicerack.redfish.Redfish: the instance.

        """
        if not password:
            password = self.management_password

        # TODO: generalize when support for additional vendors will be added.
        return RedfishDell(mgmt_fqdn, username, password, dry_run=self._dry_run)

    def alertmanager_hosts(  # pylint: disable=no-self-use
        self, target_hosts: TypeHosts, *, verbatim_hosts: bool = False
    ) -> AlertmanagerHosts:
        """Get an AlertmanagerHosts instance.

        Arguments:
            target_hosts (spicerack.typing.TypeHosts): the target hosts either as a NodeSet instance or a sequence
                of strings.
            verbatim_hosts (bool, optional): if :py:data:`True` use the hosts passed verbatim as is, if instead
                :py:data:`False`, the default, consider the given target hosts as FQDNs and extract their hostnames to
                be used in Icinga.

        Returns:
            spicerack.alertmanager.AlertmanagerHosts: AlertmanagerHosts instance.

        """
        return AlertmanagerHosts(target_hosts, verbatim_hosts=verbatim_hosts)

    def alerting_hosts(self, target_hosts: TypeHosts, *, verbatim_hosts: bool = False) -> AlertingHosts:
        """Get an AlertingHosts instance.

        Arguments:
            target_hosts (spicerack.typing.TypeHosts): the target hosts either as a NodeSet instance or a sequence
                of strings.
            verbatim_hosts (bool, optional): if :py:data:`True` use the hosts passed verbatim as is, if instead
                :py:data:`False`, the default, consider the given target hosts as FQDNs and extract their hostnames to
                be used in Icinga.

        Returns:
            spicerack.alerting.AlertingHosts: AlertingHosts instance.

        """
        return AlertingHosts(
            self.alertmanager_hosts(target_hosts, verbatim_hosts=verbatim_hosts),
            self.icinga_hosts(target_hosts, verbatim_hosts=verbatim_hosts),
        )



"""Reposync module tests."""
import os
import random
from unittest import mock

import pytest
from git import Repo
from git.objects.commit import Commit
from git.remote import PushInfo

from spicerack.remote import RemoteHosts
from spicerack.reposync import RepoSync, RepoSyncError


# pylint: disable=protected-access
class TestReposync:
    """Reposync tests class."""

    @pytest.fixture(autouse=True)
    def setup_method(self, tmp_path):
        """Setup test environment."""
        # pylint: disable=attribute-defined-outside-init
        remote_hosts = mock.MagicMock(spec_set=RemoteHosts)
        repo_dir = tmp_path / "bare_repo"
        self.bare_repo = Repo.init(repo_dir, bare=True)
        self.reposync = RepoSync(self.bare_repo, "user", remote_hosts, dry_run=False)

    def test_init(self):
        """It should be an initialised instance."""
        assert isinstance(self.reposync, RepoSync)

    def test_hexsha(self):
        """Test the hexsha property."""
        assert self.reposync.hexsha is None
        self.reposync._hexsha = "newhexsha"
        assert self.reposync.hexsha == "newhexsha"

    @pytest.mark.parametrize("commit_twice, set_env", ((False, False), (True, False), (False, True)))
    @mock.patch("spicerack.reposync.ask_confirmation")
    def test_update(self, mock_ask_confirmation, commit_twice, set_env, monkeypatch):
        """Test adding some data to the repo."""
        file_content = f"test data: {random.random()}"  # nosec
        if set_env:
            monkeypatch.setenv("SSH_AUTH_SOCK", file_content)
            assert os.environ["SSH_AUTH_SOCK"] == file_content
        with self.reposync.update("test add random data") as working_dir:
            (working_dir / "test_file.txt").write_text(file_content)
        if commit_twice:
            with self.reposync.update("test add random data") as working_dir:
                (working_dir / "test_file.txt").write_text(file_content)
        mock_ask_confirmation.assert_called_once_with(f"Ok to push changes to {self.bare_repo.common_dir}")
        assert not self.bare_repo.is_dirty()
        assert len(list(self.bare_repo.iter_commits())) == 1
        commit = next(self.bare_repo.iter_commits())
        assert commit.author.email == "noc@wikimedia.org"
        assert commit.message == "test add random data"
        # TODO: there must be a better way...
        assert f"+{file_content}" in self.bare_repo.git.show(commit.hexsha).splitlines()
        if set_env:
            monkeypatch.setenv("SSH_AUTH_SOCK", file_content)
            assert os.environ["SSH_AUTH_SOCK"] == file_content
        else:
            assert "SSH_AUTH_SOCK" not in os.environ

    @mock.patch("spicerack.reposync.ask_confirmation")
    def test_update_nochange(self, mock_ask_confirmation):
        """Test adding no data to the repo."""
        with pytest.raises(RepoSyncError):
            with self.reposync.update("test commit"):
                pass
        mock_ask_confirmation.assert_not_called()
        assert not self.bare_repo.is_dirty()
        assert len(self.bare_repo.refs) == 0

    @mock.patch("spicerack.reposync.ask_confirmation")
    def test_update_drydurn(self, mock_ask_confirmation):
        """Test adding some data to the repo."""
        self.reposync._dry_run = True
        file_content = f"test data: {random.random()}"  # nosec
        with self.reposync.update("test add random data") as working_dir:
            (working_dir / "test_file.txt").write_text(file_content)
        mock_ask_confirmation.assert_called_once_with(f"Ok to push changes to {self.bare_repo.common_dir}")
        assert not self.bare_repo.is_dirty()
        assert len(self.bare_repo.refs) == 0

    @mock.patch("spicerack.reposync.ask_confirmation")
    def test_update_bad_push(self, mock_ask_confirmation):
        """Test bad push flags."""
        file_content = f"test data: {random.random()}"  # nosec
        self.reposync._repo.create_remote("bad_remote", "http://192.0.2.1/example.git")
        with pytest.raises(RepoSyncError):
            with self.reposync.update("test add random data") as working_dir:
                (working_dir / "test_file.txt").write_text(file_content)
        mock_ask_confirmation.assert_called_once_with(f"Ok to push changes to {self.bare_repo.common_dir}")

    @mock.patch("git.Remote.push")
    @mock.patch("spicerack.reposync.ask_confirmation")
    def test_update_bad_pushinfo(self, mock_ask_confirmation, mock_push):
        """Test bad push flags."""
        self.reposync._repo.create_remote("bad_remote", "http://192.0.2.1/example.git")
        push_info = mock.MagicMock(spec_set=PushInfo)
        push_info.flags = PushInfo.ERROR
        mock_push.return_value = [push_info]

        file_content = f"test data: {random.random()}"  # nosec
        with pytest.raises(RepoSyncError):
            with self.reposync.update("test add random data") as working_dir:
                (working_dir / "test_file.txt").write_text(file_content)
        mock_ask_confirmation.assert_called_once_with(f"Ok to push changes to {self.bare_repo.common_dir}")

    def test_commit_no_hexsha(self):
        """Test push with missmatch sha1."""
        repo = mock.MagicMock(spec_set=Repo)
        commit = mock.MagicMock(spec_set=Commit)
        commit.hexsha = None
        with pytest.raises(RepoSyncError):
            self.reposync._commit(repo, "foobar")



"""Ganeti Module test."""

import json
from unittest import mock

import pytest
import requests
from ClusterShell.MsgTree import MsgTreeElem
from cumin import NodeSet

from spicerack.exceptions import SpicerackError
from spicerack.ganeti import CLUSTERS_AND_ROWS, RAPI_URL_FORMAT, Ganeti, GanetiError, GanetiRAPI, GntInstance
from spicerack.remote import Remote
from spicerack.tests import get_fixture_path


class TestGaneti:
    """Ganeti tests class."""

    def setup_method(self):
        """Setup test environment."""
        # pylint: disable=attribute-defined-outside-init
        self.remote = mock.MagicMock(spec_set=Remote)
        self.ganeti = Ganeti(username="user", password="pass", timeout=10, remote=self.remote)

        self.cluster = "ganeti01.svc.eqiad.wmnet"
        self.instance = "test.example.com"
        self.cluster_base_url = {
            cluster: RAPI_URL_FORMAT.format(cluster=cluster) + "/2" for cluster in CLUSTERS_AND_ROWS
        }
        self.base_url = self.cluster_base_url[self.cluster]
        self.instance_url = f"{self.base_url}/instances/{self.instance}"

        # load test fixtures
        with open(get_fixture_path("ganeti", "info.json"), encoding="utf-8") as info_json:
            self.info = info_json.read()
        with open(get_fixture_path("ganeti", "404.json"), encoding="utf-8") as fourohfour_json:
            self.fourohfour = fourohfour_json.read()
        with open(get_fixture_path("ganeti", "instance.json"), encoding="utf-8") as instance_json:
            self.instance_info = instance_json.read()
        with open(get_fixture_path("ganeti", "bogus.json"), encoding="utf-8") as bogus_json:
            self.bogus_data = bogus_json.read()

    def _set_requests_mock_for_instance(self, requests_mock, missing_active=False):
        """Set request mock to be 404 on all other clusters."""
        for cluster in CLUSTERS_AND_ROWS:
            if cluster == self.cluster and not missing_active:
                requests_mock.get(self.instance_url, text=self.instance_info)
            else:
                requests_mock.get(
                    f"{self.cluster_base_url[cluster]}/instances/{self.instance}",
                    text=self.fourohfour,
                    status_code=requests.codes["not_found"],
                )

    @pytest.mark.parametrize("cluster", CLUSTERS_AND_ROWS.keys())
    def test_rapi_clusters_ok(self, cluster):
        """Ganeti.rapi() should return a GanetiRAPI object."""
        assert isinstance(self.ganeti.rapi(cluster), GanetiRAPI)

    def test_rapi_clusters_missing(self):
        """If a cluster that doesn't exist is passed to Ganeti.rapi(), it should except with a SpicerackError."""
        with pytest.raises(SpicerackError, match=r"Cannot find cluster bogus cluster \(expected .*"):
            self.ganeti.rapi("bogus cluster")

    def test_rapi_master_ok(self, requests_mock):
        """The master property of a RAPI should be the hostname for the master of this cluster."""
        requests_mock.get(self.base_url + "/info", text=self.info)

        master = json.loads(self.info)["master"]
        rapi = self.ganeti.rapi(self.cluster)
        assert rapi.master == master

    def test_rapi_master_missing(self, requests_mock):
        """If a master is not specified by the uptsream API, the value of master on a RAPI should be None."""
        requests_mock.get(self.base_url + "/info", text=self.bogus_data)
        rapi = self.ganeti.rapi(self.cluster)
        assert rapi.master is None

    def test_rapi_instance_missing(self, requests_mock):
        """A RAPI object should raise a GanetiError if a requested host does not exist."""
        requests_mock.get(
            self.instance_url,
            text=self.fourohfour,
            status_code=requests.codes["not_found"],
        )
        rapi = self.ganeti.rapi(self.cluster)
        with pytest.raises(GanetiError, match=r"Non-200 from API: 404:.*"):
            rapi.fetch_instance(self.instance)

    def test_rapi_instance_request_fail(self, requests_mock):
        """A RAPI object should raise a GanetiError if a request fails."""
        rapi = self.ganeti.rapi(self.cluster)
        requests_mock.get(self.instance_url, exc=requests.exceptions.ConnectTimeout)
        with pytest.raises(GanetiError, match="Error while performing request to RAPI"):
            rapi.fetch_instance(self.instance)

    def test_rapi_instance_invalid(self, requests_mock):
        """If no mac is present in host data, fetch_instance_mac should raise a GanetiError."""
        rapi = self.ganeti.rapi(self.cluster)
        requests_mock.get(self.instance_url, text=self.bogus_data)
        with pytest.raises(GanetiError, match=""):
            rapi.fetch_instance_mac(self.instance)

    def test_rapi_instance_valid(self, requests_mock):
        """The MAC returned by RAPI.fetch_instance_mac should match the data returned by the API."""
        rapi = self.ganeti.rapi(self.cluster)
        requests_mock.get(self.instance_url, text=self.instance_info)
        mac = json.loads(self.instance_info)["nic.macs"][0]
        assert mac == rapi.fetch_instance_mac(self.instance)

    def test_fetch_cluster_for_instance_ok(self, requests_mock):
        """We should get a cluster name that the host exists for."""
        self._set_requests_mock_for_instance(requests_mock)
        assert self.ganeti.fetch_cluster_for_instance(self.instance) == self.cluster

    def test_fetch_cluster_for_instance_missing(self, requests_mock):
        """We should get an exception if the host is not found in Ganeti."""
        self._set_requests_mock_for_instance(requests_mock, missing_active=True)
        with pytest.raises(GanetiError, match="Cannot find test.example.com in any configured cluster."):
            self.ganeti.fetch_cluster_for_instance(self.instance)

    @pytest.mark.parametrize("cluster", ("", "ganeti01.svc.eqiad.wmnet"))
    def test_instance_ok(self, cluster, requests_mock):
        """It should return an instance of GntInstance for a properly configured cluster."""
        self._set_requests_mock_for_instance(requests_mock)
        requests_mock.get(self.base_url + "/info", text=self.info)
        instance = self.ganeti.instance(self.instance, cluster=cluster)
        assert isinstance(instance, GntInstance)
        assert instance.cluster == "ganeti01.svc.eqiad.wmnet"
        self.remote.query.assert_called_once_with("ganeti1.example.com")

    def test_instance_missing_master(self, requests_mock):
        """It should raise a GanetiError exception if unable to determin the instance's master to manage it."""
        self._set_requests_mock_for_instance(requests_mock)
        requests_mock.get(self.base_url + "/info", text=self.bogus_data)
        with pytest.raises(GanetiError, match="Master for cluster ganeti01.svc.eqiad.wmnet is None"):
            self.ganeti.instance(self.instance)

    @pytest.mark.parametrize("kwargs", ({}, {"timeout": 0}))
    def test_instance_shutdown(self, requests_mock, kwargs):
        """It should issue the shutdown command on the master host."""
        self._set_requests_mock_for_instance(requests_mock)
        requests_mock.get(self.base_url + "/info", text=self.info)
        instance = self.ganeti.instance(self.instance)
        instance.shutdown(**kwargs)
        timeout = kwargs["timeout"] if "timeout" in kwargs else 2
        self.remote.query.return_value.run_sync.assert_called_once_with(
            f"gnt-instance shutdown --force --timeout={timeout} test.example.com"
        )

    @pytest.mark.parametrize("kwargs", ({}, {"shutdown_timeout": 0}))
    def test_instance_remove(self, requests_mock, kwargs):
        """It should issue the remove command on the master host."""
        self._set_requests_mock_for_instance(requests_mock)
        requests_mock.get(self.base_url + "/info", text=self.info)
        instance = self.ganeti.instance(self.instance)
        instance.remove(**kwargs)
        timeout = kwargs["shutdown_timeout"] if "shutdown_timeout" in kwargs else 2
        self.remote.query.return_value.run_sync.assert_called_once_with(
            f"gnt-instance remove --shutdown-timeout={timeout} --force test.example.com"
        )

    def test_instance_add_ok(self, requests_mock):
        """It should issue the remove command on the master host."""
        self._set_requests_mock_for_instance(requests_mock)
        requests_mock.get(self.base_url + "/info", text=self.info)
        instance = self.ganeti.instance(self.instance)
        results = [
            (
                NodeSet("ganeti-master.example.com"),
                MsgTreeElem(b"creation logs", parent=MsgTreeElem()),
            )
        ]
        self.remote.query.return_value.run_sync.return_value = iter(results)

        instance.add(row="A", vcpus=2, memory=3, disk=4, link="private")

        self.remote.query.return_value.run_sync.assert_called_once_with(
            "gnt-instance add -t drbd -I hail --net 0:link=private --hypervisor-parameters=kvm:boot_order=network "
            "-o debootstrap+default --no-install -g row_A -B vcpus=2,memory=3g --disk 0:size=4g test.example.com"
        )

    @pytest.mark.parametrize(
        "kwargs, exc_message",
        (
            (
                {"row": "A", "vcpus": 1, "memory": 1, "disk": 1, "link": "invalid"},
                r"Invalid link 'invalid', expected one of: \('public', 'private', 'analytics'\)",
            ),
            (
                {
                    "row": "invalid",
                    "vcpus": 1,
                    "memory": 1,
                    "disk": 1,
                    "link": "private",
                },
                r"Invalid row 'invalid' for cluster ganeti01.svc.eqiad.wmnet, expected one of: \('A', 'B', 'C', 'D'\)",
            ),
            (
                {"row": "A", "vcpus": -1, "memory": 1, "disk": 1, "link": "private"},
                r"Invalid value '-1' for vcpus, expected positive integer.",
            ),
            (
                {"row": "A", "vcpus": 1, "memory": -1, "disk": 1, "link": "private"},
                r"Invalid value '-1' for memory, expected positive integer.",
            ),
            (
                {"row": "A", "vcpus": 1, "memory": 1, "disk": -1, "link": "private"},
                r"Invalid value '-1' for disk, expected positive integer.",
            ),
        ),
    )
    def test_instance_add_fail(self, requests_mock, kwargs, exc_message):
        """It should raise GanetiError on invalid parameters."""
        self._set_requests_mock_for_instance(requests_mock)
        requests_mock.get(self.base_url + "/info", text=self.info)
        instance = self.ganeti.instance(self.instance)
        with pytest.raises(GanetiError, match=exc_message):
            instance.add(**kwargs)

        assert not self.remote.query.return_value.run_sync.called
